SystemOrganization addCategory: #'Ma special collections'!
SystemOrganization addCategory: #'Ma special collections-Info'!
SystemOrganization addCategory: #'Ma special collections-Integer collections'!
SystemOrganization addCategory: #'Ma special collections-Squeak weak tweaks'!
SystemOrganization addCategory: #'Ma special collections-Tree'!

MaObject subclass: #MaDictionary
	instanceVariableNames: 'dicts'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma special collections-Squeak weak tweaks'!

!MaDictionary commentStamp: 'cmm 12/19/2002 23:50' prior: 0!
I am identical to a Dictionary but a lot faster when my size grows beyond 4K elements.  Squeak uses a 12-bit identity hash for its object identity, causing lots of hash collisions once you get more than 4K (2 raisedTo: 12) objects in it.  As this script demonstrates, I am faster at adding, accessing and replacing.

| sd md r |
r _ (1 to: 5500) collect: [ :e | Object new ].
r _ r shuffled.
sd _ WeakIdentityKeyDictionary new.
md _ MaWeakIdentityKeyDictionary new.

Transcript cr; show: 'time to add to sd: ', ([ r do: [ :each | sd at: each put: each ] ] timeToRun) printString.

Transcript cr; show: 'time to add to md: ', ([ r do: [ :each | md at: each put: each ] ] timeToRun) printString.

r_r shuffled.

Transcript cr; show: 'time to access to sd: ', ([ r do: [ :each | sd at: each ] ] timeToRun) printString.

Transcript cr; show: 'time to access to md: ', ([ r do: [ :each | md at: each ] ] timeToRun) printString.

r _ r shuffled.

Transcript cr; show: 'time to replace to sd: ', ([ r do: [ :each | sd at: each put: each ] ] timeToRun) printString.

Transcript cr; show: 'time to replace to md: ', ([ r do: [ :each | md at: each put: each ] ] timeToRun) printString.
!

!MaDictionary class methodsFor: 'creation' stamp: 'cmm 12/27/2002 10:09'!
new

	^self new: 8! !

!MaDictionary class methodsFor: 'creation' stamp: 'cmm 12/13/2002 14:28'!
new: anInteger

	^super new capacity: anInteger; yourself! !

!MaDictionary methodsFor: 'adding' stamp: 'cmm 12/12/2002 17:10'!
add: anAssociation

	self
		removeKey: anAssociation key
		ifAbsent: [ "do nothing" ].
	self qualifyingDict add: anAssociation! !

!MaDictionary methodsFor: 'enumerating' stamp: 'cmm 12/22/2002 18:26'!
associationsDo: aBlock

	dicts do: [ :each | each associationsDo: aBlock ]! !

!MaDictionary methodsFor: 'accessing' stamp: 'cmm 12/12/2002 16:43'!
at: anObject

	^self
		at: anObject
		ifAbsent: [ self errorKeyNotFound ]! !

!MaDictionary methodsFor: 'accessing' stamp: 'cmm 12/13/2002 14:34'!
at: anObject ifAbsent: aBlock

	| uniqueObject |
	uniqueObject _ ''.
	dicts do:
		[ :each |
		| result |
		result _ each
			at: anObject
			ifAbsent: [ uniqueObject ].
		result == uniqueObject ifFalse: [ ^result ] ].
	^aBlock value! !

!MaDictionary methodsFor: 'accessing' stamp: 'cmm 12/12/2002 16:44'!
at: key ifAbsentPut: aBlock

	^self
		at: key
		ifAbsent:
			[ self 
				at: key 
				put: aBlock value ]! !

!MaDictionary methodsFor: 'accessing' stamp: 'cmm 6/8/2004 12:51'!
at: key put: anObject

	key ifNil: [ self error: 'nil cannot be used as a key to a Dictionary.' ].
	dicts do:
		[ :each |
		(each includesKey: key) ifTrue: [ ^each at: key put: anObject ] ].
	^self qualifyingDict
		at: key
		put: anObject! !

!MaDictionary methodsFor: 'private' stamp: 'cmm 12/27/2002 09:57'!
capacity: anInteger

	| numberOfDicts |
	numberOfDicts _ anInteger // self criticalSize.
	dicts _ OrderedCollection new: numberOfDicts.
	numberOfDicts timesRepeat: [ dicts add: (self squeakDictionaryClass new: self criticalSize) ].
	anInteger \\ self criticalSize > 0
		ifTrue:
			[ dicts add: (self squeakDictionaryClass new: anInteger \\ self criticalSize) ]! !

!MaDictionary methodsFor: 'enumerating' stamp: 'cmm 12/12/2002 17:17'!
collect: aBlock 
	"Evaluate aBlock with each of my values as the argument.  Collect the
	resulting values into a collection that is like me. Answer with the new
	collection."
	| newCollection |
	newCollection _ OrderedCollection new: self size.
	self do: [:each | newCollection add: (aBlock value: each)].
	^ newCollection! !

!MaDictionary methodsFor: 'copying' stamp: 'cmm 3/24/2005 12:04'!
copy

	^ self copy postCopy! !

!MaDictionary methodsFor: 'private' stamp: 'cmm 12/27/2002 10:05'!
criticalSize
	"Since the Squeak Set allocates about 4/3rd's extra space for future growth, we have the potential to grow a little out of its efficiency range.  Compensate."

	^4096*3//4! !

!MaDictionary methodsFor: 'enumerating' stamp: 'cmm 2/5/2003 20:47'!
detect: aBlock

	^self
		detect: aBlock
		ifNone: [ self error: 'not in MaDictionary' ]! !

!MaDictionary methodsFor: 'enumerating' stamp: 'cmm 2/5/2003 20:47'!
detect: aBlock ifNone: notFoundBlock

	self do: [ :each | (aBlock value: each) ifTrue: [ ^each ] ].
	^notFoundBlock value! !

!MaDictionary methodsFor: 'enumerating' stamp: 'cmm 12/12/2002 16:58'!
do: aBlock

	dicts do: [ :each | each do: aBlock ]! !

!MaDictionary methodsFor: 'private' stamp: 'cmm 12/12/2002 16:43'!
errorKeyNotFound

	self error: 'key not found'! !

!MaDictionary methodsFor: 'private' stamp: 'cmm 12/13/2002 14:03'!
finishCopy

	dicts _ dicts collect: [ :each | each copy ]! !

!MaDictionary methodsFor: 'testing' stamp: 'cmm 12/12/2002 17:14'!
includes: anObject

	self do: [:each | anObject = each ifTrue: [^true]].
	^false! !

!MaDictionary methodsFor: 'testing' stamp: 'cmm 12/12/2002 17:14'!
includesKey: key 
	"Answer whether the receiver has a key equal to the argument, key."
	
	self at: key ifAbsent: [^false].
	^true! !

!MaDictionary methodsFor: 'accessing' stamp: 'cmm 12/12/2002 16:52'!
keys
	"Answer a Set containing the receiver's keys."
	| aSet |
	aSet _ Set new: self size.
	self keysDo: [:key | aSet add: key].
	^ aSet! !

!MaDictionary methodsFor: 'enumerating' stamp: 'cmm 12/25/2002 23:28'!
keysAndValuesDo: twoArgBlock

	dicts do: [ :each | each keysAndValuesDo: twoArgBlock ]! !

!MaDictionary methodsFor: 'removing' stamp: 'cmm 12/12/2002 17:06'!
keysAndValuesRemove: twoArgBlock

	dicts do: [ :each | each keysAndValuesRemove: twoArgBlock ]! !

!MaDictionary methodsFor: 'enumerating' stamp: 'cmm 12/12/2002 16:52'!
keysDo: aBlock

	dicts do: [ :each | each keysDo: aBlock ]! !

!MaDictionary methodsFor: 'accessing' stamp: 'cmm 12/12/2002 17:12'!
maAllKeysAtValue: anObject

	| aCollection |
	aCollection := OrderedCollection new.
	self keysAndValuesDo: [ :key :value | value = anObject ifTrue: [ aCollection add: key ] ].
	^aCollection! !

!MaDictionary methodsFor: 'accessing' stamp: 'cmm 12/13/2002 14:33'!
maAt: anObject ifPresent: oneArgBlock ifAbsent: zeroArgBlock

	| current uniqueObject |

	uniqueObject _ ''.
	^(current := self
		at: anObject
		ifAbsent: [ uniqueObject ]) == uniqueObject
			ifTrue: [ zeroArgBlock value ]
			ifFalse: [ oneArgBlock value: current ]! !

!MaDictionary methodsFor: 'testing' stamp: 'cmm 12/12/2002 17:15'!
maIsHashedCollection

	^true! !

!MaDictionary methodsFor: 'accessing' stamp: 'cmm 12/12/2002 17:16'!
occurrencesOf: anObject 
	"Answer how many of the receiver's elements are equal to anObject."

	| count |
	count _ 0.
	self do: [:each | anObject = each ifTrue: [count _ count + 1]].
	^count! !

!MaDictionary methodsFor: 'copying' stamp: 'cmm 3/24/2005 12:04'!
postCopy

	dicts _ dicts copy.
	self finishCopy! !

!MaDictionary methodsFor: 'private' stamp: 'cmm 12/12/2002 23:20'!
qualifyingDict

	^dicts
		detect: [ :each | each size < self criticalSize ]
		ifNone: [ dicts addFirst: self squeakDictionaryClass new ]! !

!MaDictionary methodsFor: 'private' stamp: 'cmm 12/12/2002 17:15'!
rehash

	dicts do: [ :each | each rehash ]! !

!MaDictionary methodsFor: 'enumerating' stamp: 'cmm 1/9/2003 00:40'!
reject: aBlock 
	"Evaluate aBlock with each of my values as the argument. Collect into a
	new dictionary, only those associations for which aBlock evaluates to
	true."

	| newCollection |
	newCollection _ self species new.
	self associationsDo: 
		[:each | 
		(aBlock value: each value) ifFalse: [newCollection add: each]].
	^newCollection! !

!MaDictionary methodsFor: 'removing' stamp: 'cmm 12/12/2002 17:04'!
remove: anObject

	self shouldNotImplement! !

!MaDictionary methodsFor: 'removing' stamp: 'cmm 12/12/2002 17:04'!
remove: anObject ifAbsent: exceptionBlock

	self shouldNotImplement! !

!MaDictionary methodsFor: 'removing' stamp: 'cmm 12/12/2002 17:05'!
removeKey: key 
	"Remove key from the receiver.
	If key is not in the receiver, notify an error."

	^ self removeKey: key ifAbsent: [self errorKeyNotFound]! !

!MaDictionary methodsFor: 'removing' stamp: 'cmm 12/13/2002 14:34'!
removeKey: key ifAbsent: aBlock

	| uniqueObject |
	uniqueObject _ ''.
	dicts do:
		[ :each |
		| removed |
		removed _ each
			removeKey: key
			ifAbsent: [ uniqueObject ].
		removed == uniqueObject ifFalse: [ ^removed ] ].
	^aBlock value! !

!MaDictionary methodsFor: 'enumerating' stamp: 'cmm 12/12/2002 17:09'!
select: aBlock 
	"Evaluate aBlock with each of my values as the argument. Collect into a
	new dictionary, only those associations for which aBlock evaluates to
	true."

	| newCollection |
	newCollection _ self species new.
	self associationsDo: 
		[:each | 
		(aBlock value: each value) ifTrue: [newCollection add: each]].
	^newCollection! !

!MaDictionary methodsFor: 'accessing' stamp: 'cmm 12/12/2002 22:38'!
size

	^dicts
		inject: 0
		into: [ :sum :each | sum + each size ]! !

!MaDictionary methodsFor: 'private' stamp: 'cmm 12/12/2002 16:48'!
squeakDictionaryClass

	^Dictionary! !

!MaDictionary methodsFor: 'accessing' stamp: 'cmm 12/12/2002 17:11'!
values
	"Answer a Collection containing the receiver's values."
	| out |
	out _ WriteStream on: (Array new: self size).
	self valuesDo: [:value | out nextPut: value].
	^ out contents! !

!MaDictionary methodsFor: 'enumerating' stamp: 'cmm 12/12/2002 16:57'!
valuesDo: aBlock

	dicts do: [ :each | each valuesDo: aBlock ]! !

MaDictionary subclass: #MaWeakIdentityKeyDictionary
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma special collections-Squeak weak tweaks'!

!MaWeakIdentityKeyDictionary methodsFor: 'private' stamp: 'cmm 1/12/2003 21:26'!
finalizeValues

	dicts do: [ :e | e finalizeValues ]! !

!MaWeakIdentityKeyDictionary methodsFor: 'private' stamp: 'cmm 12/12/2002 22:27'!
squeakDictionaryClass

	^WeakIdentityKeyDictionary! !

MaDictionary subclass: #MaWeakValueDictionary
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma special collections-Squeak weak tweaks'!

!MaWeakValueDictionary methodsFor: 'private' stamp: 'cmm 12/12/2002 22:27'!
squeakDictionaryClass

	^WeakValueDictionary! !

MaObject subclass: #MaIntervalCollection
	instanceVariableNames: 'intervals proximityThreshold'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma special collections-Integer collections'!

!MaIntervalCollection commentStamp: '<historical>' prior: 0!
This class keeps track of a collection of intervals.

As intervals are added that intersect or are within the proximityThreshold
of existing intervals, the existing interval is enlarged to include for the new interval.

If an interval is removed in the middle of a large interval such that
a "hole" is created, a new interval will be created and the other one
will be adjusted so that they will represent only the valid ranges.

For example:

	if an instance of this class has 1..10

	and we add: 8..15, then this class will have a single interval from 1..15

	Then remove 5..11..  Assuming a proximityThreshold of 1, this class will have the following two intervals:

		1..4
		12..15

Use a proximityThreshold of 0 to track the real-numbers..!

!MaIntervalCollection class methodsFor: 'creation' stamp: ''!
proximityThreshold: aMagnitude

	^self new proximityThreshold: aMagnitude! !

!MaIntervalCollection methodsFor: 'building' stamp: ''!
add: anInterval

	self
		addFrom: anInterval first
		to: anInterval last! !

!MaIntervalCollection methodsFor: 'building' stamp: 'cmm 8/18/2002 00:59'!
addFrom: from to: to

	| physicalFromIndex physicalToIndex |

	physicalFromIndex := self physicalIndexOf: from.
	physicalToIndex := self physicalIndexOf: to.
	(from = to
		and: [ self isStartOfGap: physicalFromIndex ])
			ifTrue:
				[ "User does not realize that this class deals with real-number-based
				intervals that are inclusive on the left and exclusive on the right.
				Take care of him by adding a sensible interval."
				^self addFrom: from to: to + proximityThreshold ].
	physicalFromIndex = physicalToIndex
		ifTrue:
			[ "if in a gap, add a new interval"
			((self isStartOfGap: physicalFromIndex)
				and: [ (self
					fromValue: from
					isOutsideProximityThresholdAt: physicalFromIndex)
				and:
					[ self
						toValue: to
						isOutsideProximityThresholdAt: physicalToIndex ] ])
				ifTrue:
					[ self
						privateInsertFrom: from
						to: to
						afterIndex: physicalFromIndex.
					^from ] ].
	self
		extendIntervalFrom: from
		at: physicalFromIndex
		to: to
		at: physicalToIndex.
	^from! !

!MaIntervalCollection methodsFor: 'private building' stamp: 'cmm 2/20/2002 01:17'!
bridgeFrom: physicalBridgeFromIndex to: physicalBridgeToIndex
	"Remove intervals in the receiver that are between physicalFromIndex
	and physicalToIndex which should be merged into one larger interval.
	The caller must ensure the from is the start of an interval and the to is
	the start of a gap."

	physicalBridgeToIndex - physicalBridgeFromIndex - 1
		timesRepeat: [ intervals removeAt: physicalBridgeFromIndex + 1 ]! !

!MaIntervalCollection methodsFor: 'building' stamp: 'cmm 12/20/2002 09:59'!
clear

	self
		removeFrom: self lowest
		to: self highest! !

!MaIntervalCollection methodsFor: 'private building' stamp: 'cmm 9/24/2002 12:30'!
extendGapFromValue: from at: physicalFromIndex toValue: to at: physicalToIndex

	| voidStart voidEnd |

	voidStart := (self isStartOfInterval: physicalFromIndex)
		ifTrue:
			[ "Extend to of this gap back a little."
			intervals
				at: physicalFromIndex + 1
				put: from.
			physicalFromIndex + 1 ]
		ifFalse: [ physicalFromIndex ].
	voidEnd := ((self isStartOfInterval: physicalToIndex)
		and: [ physicalToIndex > physicalFromIndex  "account for when to is an endpoint exactly on the end of an interval" ])
			ifTrue:
				[ "Account for removing exact endpoints of an existing interval."
				(intervals at: physicalToIndex + 1) + proximityThreshold = to
					ifTrue: [ physicalToIndex + 2 ]
					ifFalse:
						[ "Extend from of next gap up a little.."
						intervals
							at: physicalToIndex
							put: to.
						physicalToIndex ] ]
			ifFalse: [ physicalToIndex + 1 ].
	self
		voidFrom: voidStart
		to: voidEnd! !

!MaIntervalCollection methodsFor: 'private building' stamp: ''!
extendIntervalFrom: from at: physicalFromIndex to: to at: physicalToIndex
	
	| bridgeStart bridgeEnd |

	bridgeStart := (self isStartOfGap: physicalFromIndex)
		ifTrue:
			[ "Extend from of next interval back a little.."
			(physicalFromIndex < self physicalSize and: [ (intervals at: physicalFromIndex + 1) <= (to + proximityThreshold) ])
				ifTrue:
					[ intervals
						at: physicalFromIndex + 1 "start of next interval"
						put: from ].
			physicalFromIndex > 0
				ifTrue: [ (from - proximityThreshold) <= (intervals at: physicalFromIndex)
					ifTrue: [ physicalFromIndex - 1 ]
					ifFalse: [ physicalFromIndex + 1 ] ]
				ifFalse: [ 1 ] ]
		ifFalse:
			[ bridgeStart := physicalFromIndex ].
	bridgeEnd := (self isStartOfGap: physicalToIndex)
		ifTrue:
			[ "Extend to of this interval forward a little.."
			(physicalToIndex > 0 and: [ (intervals at: physicalToIndex) >= (from - proximityThreshold) ])
				ifTrue:
					[ intervals
						at: physicalToIndex  "end of this interval"
						put: to ].
			physicalToIndex < self physicalSize
				ifTrue: [ (to + proximityThreshold) >= (intervals at: physicalToIndex + 1)
					ifTrue: [ physicalToIndex + 2 ]
					ifFalse: [ physicalToIndex ] ]
				ifFalse: [ self physicalSize ] ]
		ifFalse:
			[ bridgeEnd := physicalToIndex + 1 ].
	self
		bridgeFrom: bridgeStart
		to: bridgeEnd! !

!MaIntervalCollection methodsFor: 'private testing' stamp: ''!
fromValue: aMagnitude isOutsideProximityThresholdAt: physicalFromIndex

	^physicalFromIndex = 0
		or: [ (aMagnitude - proximityThreshold) > (intervals at: physicalFromIndex) ]! !

!MaIntervalCollection methodsFor: 'accessing' stamp: 'cmm 8/17/2002 15:43'!
gaps

	| answer |
	answer _ OrderedCollection new.
	self gapsDo:
		[ :eachLow :eachHigh |
		answer add: (eachLow to: eachHigh) ].
	^answer! !

!MaIntervalCollection methodsFor: 'accessing' stamp: ''!
gapsDetect: twoArgBlock ifFound: foundBlock ifNone: noneBlock

	self gapsDo:
		[ :eachLow :eachHigh |
		(twoArgBlock
			value: eachLow
			value: eachHigh)
				ifTrue:
					[ ^foundBlock
						value: eachLow
						value: eachHigh ] ].
	^noneBlock value! !

!MaIntervalCollection methodsFor: '-- all --' stamp: 'cmm 8/17/2002 15:47'!
gapsDo: twoArgBlock
	"For each gap in the receivers intervals, value twoArgBlock.
	The first argument is the low of the gap, the second is the high."

	2
		to: intervals size - 2
		by: 2
		do:
			[ :eachLowIndex |
			twoArgBlock
				value: (intervals at: eachLowIndex) + proximityThreshold
				value: (intervals at: eachLowIndex + 1) - proximityThreshold ]! !

!MaIntervalCollection methodsFor: 'testing' stamp: 'cmm 8/25/2002 20:45'!
hasGaps

	^intervals size > 2! !

!MaIntervalCollection methodsFor: 'accessing' stamp: ''!
highest

	^intervals isEmpty ifTrue: [ self error: self printString, ' is empty.' ] ifFalse: [ intervals last ]! !

!MaIntervalCollection methodsFor: 'testing' stamp: ''!
includes: aMagnitude

	^(self indexOf: aMagnitude) > 0! !

!MaIntervalCollection methodsFor: 'private' stamp: 'cmm 3/24/2005 12:02'!
independentCopy

	^ self copy postIndependentCopy! !

!MaIntervalCollection methodsFor: 'accessing' stamp: ''!
indexOf: aMagnitude
	"Answer the index of the interval which contains aMagnitude.
	If aMagnitude is not part of any interval in the receiver, answer 0."

	| index |

	index := self physicalIndexOf: aMagnitude.
	^((self isStartOfInterval: index)
		and: [ (intervals at: index + 1) > aMagnitude ])
			ifTrue:
				[ self logicalIndexFor: index ]
			ifFalse:
				[ 0 ]! !

!MaIntervalCollection methodsFor: 'private' stamp: 'cmm 7/17/2002 22:53'!
initialize

	super initialize.

	"The intervals is stored in sorted order.  Each odd is the low of an interval, the
	even is the high."
	intervals := OrderedCollection new: 2.

	"proximityThreshold is a value used to determine how close the
	To end of an interval needs to be to the From end of the next
	interval before they will be merged into a single interval.  The
	default is 0 (no merging)."
	proximityThreshold := 0! !

!MaIntervalCollection methodsFor: 'accessing' stamp: 'cmm 8/17/2002 15:40'!
intervals
	"This needed only for the MaPersistentObjectsApp.  Normally, we don't need this, but to efficiently update the Magma 'memory' on disk, it is best to have direct access to this."

	^intervals
! !

!MaIntervalCollection methodsFor: 'actions' stamp: ''!
intervalsDo: twoArgBlock
	"For each interval in the receivers intervals, value twoArgBlock.
	The first argument is the low of the interval, the second is the high."

	1
		to: intervals size - 1
		by: 2
		do:
			[ :eachLowIndex |
			twoArgBlock
				value: (intervals at: eachLowIndex)
				value: (intervals at: eachLowIndex + 1) ]! !

!MaIntervalCollection methodsFor: 'testing' stamp: ''!
isEmpty

	^intervals isEmpty! !

!MaIntervalCollection methodsFor: 'private testing' stamp: ''!
isStartOfGap: physicalIndex

	^physicalIndex even! !

!MaIntervalCollection methodsFor: 'private testing' stamp: ''!
isStartOfInterval: physicalIndex

	^physicalIndex odd! !

!MaIntervalCollection methodsFor: 'private' stamp: ''!
logicalIndexFor: physicalIndex

	^(physicalIndex + 1) // 2! !

!MaIntervalCollection methodsFor: 'accessing' stamp: 'cmm 8/29/2002 22:45'!
lowest

	^intervals first! !

!MaIntervalCollection methodsFor: 'accessing' stamp: ''!
maDetect: twoArgBlock ifFound: foundBlock ifNone: noneBlock 
	self intervalsDo: 
			[:eachLow :eachHigh | 
			(twoArgBlock value: eachLow value: eachHigh) 
				ifTrue: [^foundBlock value: eachLow value: eachHigh]].
	^noneBlock value! !

!MaIntervalCollection methodsFor: 'printing' stamp: 'cmm 2/5/2003 16:33'!
maPrintAbbreviatedOn: aStream

	1
		to: intervals size
		do:
			[ :index |
			aStream nextPutAll: (intervals at: index) printString.
			index odd ifTrue: [ aStream nextPutAll: '..' ] ifFalse: [ aStream nextPutAll: '  ' ] ]
! !

!MaIntervalCollection methodsFor: 'testing' stamp: ''!
notEmpty

	^self isEmpty not
! !

!MaIntervalCollection methodsFor: 'private' stamp: ''!
physicalIndexFor: logicalIndex

	^(logicalIndex * 2 - 1) max: 0! !

!MaIntervalCollection methodsFor: 'private' stamp: ''!
physicalIndexOf: aMagnitude

	| index |

	index := intervals size.
	[ index > 0 and: [ (intervals at: index) >= aMagnitude ] ] whileTrue: [ index := index - 1 ].
	^index! !

!MaIntervalCollection methodsFor: 'private' stamp: ''!
physicalSize

	^intervals size! !

!MaIntervalCollection methodsFor: 'private creation' stamp: 'cmm 7/1/2004 10:49'!
postCopy

	intervals := intervals copy! !

!MaIntervalCollection methodsFor: 'private' stamp: 'cmm 3/24/2005 12:02'!
postIndependentCopy

	intervals _ intervals copy! !

!MaIntervalCollection methodsFor: 'private building' stamp: ''!
privateInsertFrom: low to: high afterIndex: physicalIndex

	intervals
		add: high
		afterIndex: physicalIndex.
	intervals
		add: low
		afterIndex: physicalIndex! !

!MaIntervalCollection methodsFor: 'private building' stamp: ''!
proximityThreshold: aMagnitude

	proximityThreshold := aMagnitude! !

!MaIntervalCollection methodsFor: 'building' stamp: ''!
remove: anInterval

	^self
		removeFrom: anInterval first
		to: anInterval last! !

!MaIntervalCollection methodsFor: 'building' stamp: 'cmm 8/18/2002 00:59'!
removeFrom: from to: to

	| physicalFromIndex physicalToIndex |

	from = to
		ifTrue:
			[ "User does not realize that this class deals with real-number-based
			intervals that are inclusive on the left and exclusive on the right.
			Take care of him by removing a sensible interval."
			^self
				removeFrom: from
				to: to + proximityThreshold ].
	physicalFromIndex := self physicalIndexOf: from.
	physicalToIndex := self physicalIndexOf: to.
	physicalFromIndex = physicalToIndex
		ifTrue:
			[ "If completely inside an interval, we must add a new interval to break it up"
			((self isStartOfInterval: physicalFromIndex)
				and: [ (intervals at: physicalFromIndex + 1) > to ])
					ifTrue:
						[ self
							privateInsertFrom: from - proximityThreshold
							to: to + proximityThreshold
							afterIndex: physicalFromIndex.
						^from ] ].
	self
		extendGapFromValue: from - proximityThreshold
		at: physicalFromIndex
		toValue: to + proximityThreshold
		at: physicalToIndex.
	^from! !

!MaIntervalCollection methodsFor: 'accessing' stamp: ''!
size

	^intervals size / 2! !

!MaIntervalCollection methodsFor: 'private testing' stamp: ''!
toValue: aMagnitude isOutsideProximityThresholdAt: physicalToIndex

	^physicalToIndex = self physicalSize
		or: [ (aMagnitude + proximityThreshold) < (intervals at: physicalToIndex + 1) ]! !

!MaIntervalCollection methodsFor: 'private building' stamp: 'cmm 7/3/2002 00:13'!
voidFrom: voidStart to: voidEnd

	voidEnd - voidStart - 1
		timesRepeat: [ intervals removeAt: voidStart + 1 ]! !

MaObject subclass: #MaPermutator
	instanceVariableNames: 'arguments'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma special collections'!

!MaPermutator class methodsFor: 'actions' stamp: ''!
permutationsOf: anArrayOfArrays do: aBlock
	"For each combination of elements in anArrayOfArray's, value aBlock
	with the arguments of each combination of elements.  There must be
	as many arguments in aBlock as there are elements in the outer array
	of Array's.

	Note, this also works with any sequenceable collection of sequenceable
	collections."

	| numberOfCombinations args |

	numberOfCombinations :=
		anArrayOfArrays
			inject: 1
			into: [ :prod :each | each size * prod ].
	args := Array new: anArrayOfArrays size.
	1
		to: numberOfCombinations
		do:
			[ :elementIndex |
			1
				to: anArrayOfArrays size
				do:
					[ :innerIndex | | eachInnerArray radix |
					eachInnerArray := anArrayOfArrays at: innerIndex.
					radix := 1.
					1
						to: innerIndex - 1
						do: [ :eachOrder | radix := radix * (anArrayOfArrays at: eachOrder) size ].
					args
						at: innerIndex
						put: (eachInnerArray at: (elementIndex // radix \\ eachInnerArray size + 1)) ].
			aBlock valueWithArguments: args ]
! !

!MaPermutator methodsFor: 'building' stamp: ''!
addArguments: aSequenceableCollection

	^self arguments add: aSequenceableCollection
! !

!MaPermutator methodsFor: 'accessing' stamp: ''!
arguments

	^arguments! !

!MaPermutator methodsFor: 'building' stamp: ''!
arguments: aSequenceableCollectionOfSequenceableCollections

	arguments := aSequenceableCollectionOfSequenceableCollections
! !

!MaPermutator methodsFor: 'private initializing' stamp: 'cmm 12/24/2002 15:51'!
initialize
	super initialize.
	arguments := OrderedCollection new! !

!MaPermutator methodsFor: 'actions' stamp: ''!
permutationsDo: oneArgBlock
	"For each combination of elements in the receivers args, value aBlock
	with the arguments of each combination of elements.  The argument to
	oneArgBlock is the Array of parameters for each combination."

	| args |

	args := Array new: self arguments size.
	1
		to: self size
		do:
			[ :elementIndex |
			1
				to: self arguments size
				do:
					[ :innerIndex | | eachInnerArray radix |
					eachInnerArray := self arguments at: innerIndex.
					radix := 1.
					1
						to: innerIndex - 1
						do: [ :eachOrder | radix := radix * (self arguments at: eachOrder) size ].
					args
						at: innerIndex
						put: (eachInnerArray at: (elementIndex // radix \\ eachInnerArray size + 1)) ].
			oneArgBlock value: args ]
! !

!MaPermutator methodsFor: 'actions' stamp: ''!
permutationsWhere: conditionBlock do: oneArgBlock
	"Same as permutationsDo: except the conditionBlock must also be
	satisfied for aBlock to run.  This is only for mildly better looking code
	since a case statement could achieve the same effect."

	| args |

	args := Array new: self arguments size.
	1
		to: self size
		do:
			[ :elementIndex |
			1
				to: self arguments size
				do:
					[ :innerIndex | | eachInnerArray radix |
					eachInnerArray := self arguments at: innerIndex.
					radix := 1.
					1
						to: innerIndex - 1
						do: [ :eachOrder | radix := radix * (self arguments at: eachOrder) size ].
					args
						at: innerIndex
						put: (eachInnerArray at: (elementIndex // radix \\ eachInnerArray size + 1)) ].
			(conditionBlock value: args) ifTrue: [ oneArgBlock value: args ] ]
! !

!MaPermutator methodsFor: 'accessing' stamp: ''!
size

	^self arguments
		inject: 1
		into: [ :prod :each | each size * prod ]
! !

MaObject subclass: #MaTree
	instanceVariableNames: 'parents'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma special collections-Tree'!

!MaTree class methodsFor: 'creation' stamp: ''!
new

	^super new initialize
! !

!MaTree methodsFor: 'building' stamp: ''!
add: anObject
	"Adds anObject as a main parent in the receiver."

	self
		add: anObject
		asChildOf: nil.
	^anObject
! !

!MaTree methodsFor: 'building' stamp: 'cmm 2/12/2004 00:02'!
add: anObject asChildOf: aParentObject

	parents at: anObject put: aParentObject.
	^anObject! !

!MaTree methodsFor: 'building' stamp: ''!
addAll: aCollection

	"Adds all in aCollection as a main parents in the receiver."

	aCollection do: [ :each | self add: each ]
! !

!MaTree methodsFor: 'building' stamp: 'cmm 6/10/2003 22:59'!
addPath: anOrderedCollection

	"Adds each object in anOrderedCollection as a new path in
	the receiver.  The first element in anOrderedCollection is the
	main parent, and each subsequent element is a child of the
	previous.  Any parts of the path that need created will be."

	^self
		addPath: anOrderedCollection
		to: nil
! !

!MaTree methodsFor: 'building' stamp: ''!
addPath: anOrderedCollection to: parentObject
	"Adds each object in anOrderedCollection as a new path in
	the receiver.  The first element in anOrderedCollection is the
	main parent, and each subsequent element is a child of the
	previous.  If part of the path already exists, there it is
	replaced (though they should be equivalent anyway)."

	anOrderedCollection
		inject: parentObject
		into:
			[ :parent :each |
			self
				add: each
				asChildOf: parent.
			each ].
	^anOrderedCollection
! !

!MaTree methodsFor: 'building' stamp: ''!
addTree: aMaTree

	^self addTree: aMaTree to: nil
! !

!MaTree methodsFor: 'building' stamp: 'cmm 12/24/2002 16:08'!
addTree: aMaTree to: anObject

	(self includes: anObject) ifFalse: [ self error: anObject printString, ' is not part of ', self printString ].
	aMaTree withParentsDo: [ :each :eachParent | self add: each asChildOf: eachParent ].
	^aMaTree
! !

!MaTree methodsFor: 'enumerating' stamp: ''!
allLeafsDo: aBlock
	"Value aBlock with each element in the receiver that has no children."

	| children |

	children := self children.
	parents keysAndValuesDo:
		[ :eachChild :eachParent |
		(children includesKey: eachChild) ifFalse: [ aBlock value: eachChild ] ]
! !

!MaTree methodsFor: 'accessing' stamp: ''!
allMembers
	"Returns a Set of all the members of the hierarchy"

	^parents keys! !

!MaTree methodsFor: 'accessing' stamp: 'cmm 6/10/2003 23:04'!
allMembersDo: aBlock
	"Returns a Set of all the members of the hierarchy"

	^parents keysDo: aBlock! !

!MaTree methodsFor: 'accessing' stamp: 'cmm 1/23/2002 14:30'!
children

	| children |

	children := Dictionary new: parents size.
	parents keysAndValuesDo:
		[ :eachChild :eachParent |
		(children
			at: eachParent
			ifAbsentPut: [ OrderedCollection new ])
				add: eachChild ].
	^children
! !

!MaTree methodsFor: 'accessing' stamp: ''!
childrenOf: anObject 
	^parents maAllKeysAtValue: anObject! !

!MaTree methodsFor: 'accessing' stamp: ''!
collect: aBlock
	"Returns a new tree whose nodes have all been operated on by aBlock."

	^self
		collect: aBlock
		into: self class new
! !

!MaTree methodsFor: 'private' stamp: 'cmm 1/23/2002 14:30'!
collect: aBlock into: aMaTree 
	"Since some objects may share a parent, we use the map to
	make sure we only value each parent just once, otherwise we
	could have some identity problems."

	| map |
	map := Dictionary new.
	parents keysAndValuesDo: 
			[:eachKey :eachValue | 
			eachValue isNil 
				ifTrue: [aMaTree add: (map at: eachKey ifAbsentPut: [aBlock value: eachKey])]
				ifFalse: 
					[aMaTree add: (map at: eachKey ifAbsentPut: [aBlock value: eachKey])
						asChildOf: (map at: eachValue ifAbsentPut: [aBlock value: eachValue])]].
	^aMaTree! !

!MaTree methodsFor: 'accessing' stamp: ''!
conform: aBlock 
	parents keysDo: [:each | (aBlock value: each) ifFalse: [^false]].
	^true! !

!MaTree methodsFor: 'building' stamp: 'cmm 3/24/2005 11:06'!
copy

	^ super copy
		parents: parents ;
		yourself! !

!MaTree methodsFor: 'enumerating' stamp: ''!
depthFirstDo: twoArgBlock
	"Value twoArgBlock with each object in the receiver and its parent in the order
	of depthFirst."

	self roots do:
		[ :each |
		twoArgBlock
			value: each
			value: nil.
		self
			depthFirstDo: twoArgBlock
			at: each ]
! !

!MaTree methodsFor: 'enumerating' stamp: ''!
depthFirstDo: twoArgBlock at: anObject

	(self childrenOf: anObject) do:
		[ :each |
		twoArgBlock
			value: each
			value: anObject.
		self
			depthFirstDo: twoArgBlock
			at: each ]
! !

!MaTree methodsFor: 'accessing' stamp: ''!
depthOf: anObject
	"Returns the number of parents anObject has."

	| depth prevChild |

	(self includes: anObject)
		ifFalse:
			[ self error: self printString , ' does not include ' , anObject printString ].
	depth := 0.
	prevChild := anObject.
	[ prevChild isNil ] whileFalse:
		[ prevChild := self parentOf: prevChild.
		depth := depth + 1 ].
	^depth
! !

!MaTree methodsFor: 'accessing' stamp: 'cmm 6/10/2003 23:02'!
detect: aBlock

	^parents detect: aBlock
! !

!MaTree methodsFor: 'accessing' stamp: 'cmm 6/10/2003 23:02'!
detect: aBlock ifNone: exceptionBlock

	^parents
		detect: aBlock
		ifNone: exceptionBlock value
! !

!MaTree methodsFor: 'enumerating' stamp: 'cmm 12/24/2002 16:07'!
do: oneArgBlock

	parents keysDo: oneArgBlock! !

!MaTree methodsFor: 'testing' stamp: 'cmm 7/25/2002 19:43'!
hasChildFor: anObject 
	"Returns true if anObject has any children.  Signal an error
	if anObject is not part of the hierarchy."

	^parents includes: anObject! !

!MaTree methodsFor: 'testing' stamp: ''!
hasParentFor: anObject
	"Returns true if anObject has a parent"

	^(self parentOf: anObject) notNil
! !

!MaTree methodsFor: 'testing' stamp: ''!
includes: anObject 
	^parents includesKey: anObject! !

!MaTree methodsFor: 'private' stamp: 'cmm 3/24/2005 11:07'!
initialize
	super initialize.
	self parents: Dictionary new! !

!MaTree methodsFor: 'enumerating' stamp: 'cmm 6/10/2003 23:02'!
maDetect: aBlock ifFound: foundBlock ifNone: exceptionBlock 
	^parents
		maDetect: aBlock
		ifFound: foundBlock
		ifNone: exceptionBlock value! !

!MaTree methodsFor: 'building' stamp: 'cmm 6/10/2003 23:00'!
move: anObject to: newParentObject 
	^parents at: anObject put: newParentObject! !

!MaTree methodsFor: 'accessing' stamp: ''!
nodesInPathFrom: someChildObject to: someParentObject
	"Returns a collection of nodes between someChildObject
	and someParentObject."

	|answer node|

	answer := OrderedCollection new.
	self
		nodesInPathFrom: someChildObject
		to: someParentObject
		do: [ :each | answer add: each ].
	^answer
! !

!MaTree methodsFor: 'enumerating' stamp: ''!
nodesInPathFrom: someChildObject to: someParentObject do: oneArgBlock
	"For each node in the tree from someChildObject to someParentObject, value
	oneArgBlock.  The highest level parent (nil) will not be used as an argument
	to value oneArgBlock."

	| node |

	node := someChildObject.
	oneArgBlock value: someChildObject.
	[ node isNil or: [ node = someParentObject ] ] whileFalse:
		[ oneArgBlock value: node.
		node := self parentOf: node ].
	someParentObject notNil ifTrue: [ oneArgBlock value: someParentObject ]
! !

!MaTree methodsFor: 'accessing' stamp: ''!
parentOf: anObject 
	^parents at: anObject! !

!MaTree methodsFor: 'private' stamp: 'cmm 3/24/2005 11:07'!
parents: aDictionary

	parents _ aDictionary! !

!MaTree methodsFor: 'accessing' stamp: ''!
pathFrom: anObject
	"Returns a collection of the objects that are contained in
	the path to anObject ordered from an Object to the top of the hierarchy."

	^self
		nodesInPathFrom: anObject
		to: nil
! !

!MaTree methodsFor: 'enumerating' stamp: ''!
pathFrom: anObject do: aBlock
	"Value aBlock with the objects from anObject to its highest level parent."

	^self
		nodesInPathFrom: anObject
		to: nil
		do: aBlock
! !

!MaTree methodsFor: 'accessing' stamp: ''!
pathTo: anObject
	"Returns a collection of the objects that are contained in
	the path to anObject, ordered from the the top
	of the hierarchy to anObject."

	^(self pathFrom: anObject) reverse
! !

!MaTree methodsFor: 'testing' stamp: ''!
pathTo: anObject includes: anotherObject
	"Returns a boolean.  Example:  in the Smalltalk
	class hierarchy, pathTo: Integer includes: Magnitude would
	return true.  We use pathFrom in the implementation because
	it is faster."

	^(self pathFrom: anObject) includes: anotherObject
! !

!MaTree methodsFor: 'printing' stamp: 'cmm 12/24/2002 16:08'!
printOn: aStream

	super printOn: aStream.
	self depthFirstDo:
		[ :eachChild :eachParent |
		aStream
			cr;
			tab: (self depthOf: eachChild);
			maPrintAbbreviated: eachChild ]
! !

!MaTree methodsFor: 'accessing' stamp: 'cmm 7/1/2004 00:34'!
reject: aBlock 
	"Returns a new Tree with all branches removed that
	meet aBlock criteria.  See method comment for more
	info."

	^self reject: [:each | aBlock value: each ]! !

!MaTree methodsFor: 'building' stamp: 'cmm 7/25/2002 19:42'!
remove: anObject
	"Removes anObject and all of its children."

	^self
		remove: anObject
		ifAbsent: [ self error: anObject printString, ' was not found in ', self printString ]
! !

!MaTree methodsFor: 'building' stamp: ''!
remove: anObject ifAbsent: oneArgBlock 
	"Answers the collection of removed objects removed after removing
	anObject after removing it and all of its children."

	| tree |
	(tree := self treeAt: anObject) allMembers 
		do: [:each | parents removeKey: each]! !

!MaTree methodsFor: 'accessing' stamp: 'cmm 6/10/2003 23:08'!
roots
	"Answers a collection of all roots in this tree."

	^self childrenOf: nil
! !

!MaTree methodsFor: 'accessing' stamp: 'cmm 6/10/2003 23:05'!
select: aBlock
	"Returns a new Tree with all branches removed that
	meet aBlock criteria.  See method comment for more
	info."

	| newTree |

	newTree := self class new.
	self allMembersDo:
		[ : each |
		(aBlock value: each) ifTrue: [ newTree addPath: (self pathTo: each) ] ].
	^newTree! !

!MaTree methodsFor: 'accessing' stamp: ''!
size
	^parents size! !

!MaTree methodsFor: 'accessing' stamp: 'cmm 7/25/2002 19:43'!
treeAt: aNode

	"Return a new instance of the receiver which has a single main
	parent aNode."

	^self
		treeAt: aNode
		ifAbsent: [ self error: aNode printString, ' was not found in ', self printString ]
! !

!MaTree methodsFor: 'accessing' stamp: ''!
treeAt: anObject ifAbsent: aBlock
	"Return a new instance of the receiver which has a single main
	parent anObject."

	| newTree |

	(self includes: anObject) ifFalse: [ ^aBlock value ].
	newTree := self class new.
	self
		depthFirstDo:
			[ :each :eachParent |
			newTree
				add: each
				asChildOf: eachParent ]
		at: anObject.
	^newTree
! !

!MaTree methodsFor: 'enumerating' stamp: 'cmm 12/24/2002 16:09'!
withParentsDo: twoArgBlock

	parents keysAndValuesDo: twoArgBlock! !

PackageInfo subclass: #MaSpecialCollectionsPackageInfo
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma special collections-Info'!

!MaSpecialCollectionsPackageInfo methodsFor: 'as yet unclassified' stamp: 'cmm 1/29/2003 23:33'!
prerequisitePackages

	^#('Ma base additions')! !

!MaSpecialCollectionsPackageInfo methodsFor: 'as yet unclassified' stamp: 'cmm 1/30/2003 00:38'!
readmeText

	^'I consist of some really special-purpose Collection classes that seem to be generally useful.

MaWeakIdentityKeyDictionary and MaWeakValueDictionary wrap a collection of their Smalltalk counterparts and keeps them all under 12-bits in size, thus minimizing the number of collisions due to Squeaks 12-bits identity hash.

MaTree is one of my first attempts at a Smalltalk collection.  It''s API compatible with the other collections.  Example:

	MaTree new
		add: ''Great Grandma'';
		add: ''Grandma'' asChildOf: ''Great Grandma'';
		add: ''Mom'' asChildOf: ''Grandma'';
		yourself

MaIntervalCollection is one of the most interesting collections.  It can efficiently track interval ranges and merge them when they get within their proximityThreshold.  This is useful for recording occupied space in a large range or, perhaps grouping entities based on their proximity to each other.  For example:

	(MaIntervalCollection proximityThreshold: 1)
		addFrom: 5 to: 10;
		addFrom: 15 to: 20;
		addFrom: 25 to: 30;
		addFrom: 21 to: 24;
		yourself'
! !
